/*
 * screen.c --- screen package.
 *
 * Copyright (c) 1997-2002, 2015 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of Natch.
 *
 * Natch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Natch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include	"common.h"

#ifdef HAVE_LIBNCURSES

#include	"screen.h"
#include        "main.h"
#include	"langues.h"
#include	"hash.h"
#include	"pcpj.h"
#include	"pcpjtool.h"

#include	<string.h>	/* for strlen(...) */
#include	<ctype.h>	/* for tolower(...) */
#include	<stdlib.h>	/* For exit(...) */

/*----------------------------------------------------------*/
#define		LINE_SIZE		64
#define		NB_LINES		9

#define		STOP_X			8
#define		STOP_Y			4
/*----------------------------------------------------------*/
typedef	char		line_t[ LINE_SIZE ];
static line_t		Lines[ NB_LINES ];
static int		CurLine;
static int		CurPosition;
static int		FirstLine;
static const char*	StrCoupsRestants;

int ScreenKeyboardCounter = 0;

/*----------------------------------------------------------*/

static unsigned int MaxCol;

/* Maximum size for showing one move. */
static int MaxWhite, MaxBlack;

/* Immobile pieces */
static const piece_t* Immobile_pieces;
static unsigned int Nb_immobile_pieces;

/*----------------------------------------------------------*/

int Screen_Key;                 /* Key typed by the user to enter function
                                 * ScreenKeyBoard */

typedef unsigned long	attribut_t;

#define		ATTR_WHITE_ON_BLACK	COLOR_PAIR(0)

#define		ATTR_P_WHITE_ON_WHITE	COLOR_PAIR(1)
#define		ATTR_P_WHITE_ON_BLACK	COLOR_PAIR(2)
#define		ATTR_P_BLACK_ON_WHITE	COLOR_PAIR(3)
#define		ATTR_P_BLACK_ON_BLACK	COLOR_PAIR(4)

#define		ATTR_RED_ON_BLACK	COLOR_PAIR(5)
#define		ATTR_BROWN_ON_BLACK	COLOR_PAIR(6)

#include	<curses.h>

/*----------------------------------------------------------*/

static void
screenSetSize(int row __attribute__ ((unused)), int col)
{
    MaxCol = col;

    MaxWhite = col > 40 ? 40 : col;
    MaxBlack = col > 40 ? col - 40 : 0;
}

static void
screenWriteImmobileAssoc()
{
    unsigned int i;
    for (i = 0; i < Nb_immobile_pieces; i++)
    {
        screenWriteAssoc(&Immobile_pieces[i]);
    }
}

static void
screenPosition(
    const piece_index_t* board,
    const piece_type_t*	types,
    int			number_of_half_moves)
{
    int			i, j;
    piece_index_t	piece_index;
    int			nbBlanc, nbNoir;
    char		s[60];
    attribut_t		attr;
    const char* const	initialePieces = langGetString( eStrInitialOfPieces );
    int                 nbMoves;
    const char*         half;

    nbBlanc = nbNoir = 0;
    for ( i = 7; i >= 0; i-- )
    {
        for ( j = 0; j < 8; j++ )
        {
            piece_index = board[ i * 16 + j ];
            if ( piece_index != 0 )
            {
                if ( is_white( piece_index ) )
                {
                    nbBlanc++;
                    if ( ( i + j ) & 1 )	attr = ATTR_P_WHITE_ON_WHITE;
                    else			attr = ATTR_P_WHITE_ON_BLACK;
                }
                else
                {
                    nbNoir++;
                    if ( ( i + j ) & 1 )	attr = ATTR_P_BLACK_ON_WHITE;
                    else			attr = ATTR_P_BLACK_ON_BLACK;
                }
                mvaddch(7 - i, j, attr | initialePieces[types[piece_index]]);
            }
            else	/* case vide */
            {
                if ( ( i + j ) & 1 )	attr = ATTR_P_WHITE_ON_WHITE;
                else			attr = ATTR_P_WHITE_ON_BLACK;
                mvaddch(7 - i, j, attr | ' ');
            }
        }
    }

    sprintf( s, "%s %s", PACKAGE, VERSION );
    if ( strlen( s ) > 9 )
    {
        s[ 9 ] = 0 ;
    }
    attrset(ATTR_WHITE_ON_BLACK);
    mvaddstr(0, 8, s);

    sprintf( s, "%s %s", PACKAGE, VERSION );
    if ( strlen( s ) > 9 )
    {
        mvaddstr(1, 9, s[9] == '.' ? &s[10] : &s[9]);
    }

    nbMoves = number_of_half_moves / 2;
    half = langGetString(
        (number_of_half_moves & 1) == 0 ? eStrDotZero : eStrDotFive);

    sprintf(s, "%d%s %s", nbMoves, half, langGetString(eStrMoves));
    mvaddstr(8, 0, s);
    sprintf(s, "%2d + %d", nbBlanc, nbNoir);
    mvaddstr(7, 9, s);

    refresh();

    for ( i = 0; i < NB_LINES; i++ )
    {
        Lines[ i ][ LINE_SIZE - 1 ] = '\0';
    }
}

void
screenWriteAssoc(const piece_t* p)
{
    int		i;
    char	s[120];
    int		x, y;
    int		total_number_of_moves = 0;
    bool_t      is_white = p->index < INDEX_ROI_NOIR;
    int         max = is_white ? MaxWhite : MaxBlack;

    for (i=0; i<p->nbDestination; i++)
    {
        total_number_of_moves += p->destinations[i].distance;
    }

    if (total_number_of_moves >= 10)
    {
        s[0] = '0' + total_number_of_moves / 10;
        s[1] = '0' + total_number_of_moves - 10;
    }
    else if (total_number_of_moves == 0)
    {
        s[0] = ' ';
        s[1] = ' ';
    }
    else
    {
        s[0] = ' ';
        s[1] = '0' + total_number_of_moves;
    }
    s[2] = ' ';
    getOneAssociationNoEnd(p, &s[3]);
    s[max] = '\0';

    x = is_white ? 0 : 40;
    y = 9 + (p->index & 0x0F);

    attrset(ATTR_WHITE_ON_BLACK);

    if (mvaddstr(y, x, s) != ERR)
    {
        i = strlen(s);
        if (p->pieceCapturante != PIECE_PAS_CAPTUREE)
        {
            piece_t* capturer = TabIndexPiece[p->pieceCapturante];
            attrset(ATTR_BROWN_ON_BLACK);
            addch('*');
            addch(InitialePieces[capturer->typePiece]);
            addstr(square_to_string(capturer->caseInitiale));
            i += 4;
        }
        for (; i<max; i++) addch(' ');
    }
}

void
screenInit(
    const piece_index_t* board,
    const piece_type_t*	types,
    unsigned int	number_of_half_moves,
    const piece_t*	pieces,
    unsigned int	nb_pieces)
{
    int row, col;

    initscr();
    getmaxyx(stdscr, row, col);
    screenSetSize(row, col);

    start_color();

    cbreak();
    noecho();
    nodelay(stdscr, TRUE);
    nonl();
    keypad(stdscr, TRUE);
    intrflush(stdscr, FALSE);

    /* COLOR_YELLOW is used because it gives BROWN on i386 !
     * See documentation in curs_color(3X).
     */
    init_pair(1, COLOR_RED,    COLOR_WHITE);	/* White on white square */
    init_pair(2, COLOR_RED,    COLOR_YELLOW);	/* White on black square */
    init_pair(3, COLOR_BLACK,  COLOR_WHITE);	/* Black on white square */
    init_pair(4, COLOR_BLACK,  COLOR_YELLOW);	/* Black on black square */
    init_pair(5, COLOR_RED,    COLOR_BLACK);	/* Red on black background */
    init_pair(6, COLOR_YELLOW, COLOR_BLACK);	/* Brown on black background */

    /* Set invisible cursor : if this doesn't work : the result of the function
       will be ERR, we just ignore it. */
    curs_set( 0 );

    erase();
    move(0, 0);
    screenPosition( board, types, number_of_half_moves );

    Immobile_pieces = pieces;
    Nb_immobile_pieces = nb_pieces;
    screenWriteImmobileAssoc();

    StrCoupsRestants = langGetString( eStrCoupsRestants );

    screenCoup( 0, 0, NULL );	/* Affiche "Recherche position" + refresh */
}

void
screen(
    const piece_t*	pieces,
    unsigned int	nb_pieces,
    unsigned long	nb_position,
    unsigned int	white_spare_moves,
    unsigned int	black_spare_moves)
{
    unsigned int i, l;
    char s[20];

    for( i = 0; i < nb_pieces; i++ )
    {
        screenWriteAssoc( &pieces[ i ] );
    }

    attrset(ATTR_WHITE_ON_BLACK);

    sprintf( s, "%lu", nb_position );
    mvaddstr(2, 9, s);

    sprintf( s, StrCoupsRestants, white_spare_moves, black_spare_moves );
    l = strlen(s);
    while (l < 9)
    {
        s[l] = ' ';
        l++;
    }
    if (l > 9)	l = 9;
    s[l] = '\0';
    mvaddstr(3, 8, s);

    refresh();
}

void
screenCoup(int i, int nb, const UCHAR* coup)
{
    char	s[60];

    attrset(ATTR_WHITE_ON_BLACK);
    if (i == 0)
    {
        mvaddstr(5, 8, langGetString(eStrRecherche));
        mvaddstr(6, 8, langGetString(eStrPosition));
    }
    else
    {
        sprintf(s, " %2d / %d  ", i, nb);
        s[9] = '\0';		/* Il peut y avoir 1 caracteres de trop. */
        mvaddstr(5, 8, s);
        (void)move_to_string(coup, &s[1]);
        s[0] = ' ';
        s[9] = '\0';		/* Le coup a 9 caracteres donc 1 de trop */
        mvaddstr(6, 8, s);
    }

    refresh();
}

/*------------------------------------------------------------------*/

static void
screenPrintLines()
{
    int i;
    char s[160];

    attrset(ATTR_WHITE_ON_BLACK);
    for (i=0; i<NB_LINES; i++)
    {
        strcpy(s, Lines[FirstLine]);
        if (strlen(s) > MaxCol - 17) s[MaxCol - 17] = '\0';
        mvaddstr(i, 17, s);
        FirstLine++;
        if (FirstLine == NB_LINES)	FirstLine = 0;
    }

    refresh();
}

static void
screenIncLine()
{
    CurPosition = 0;
    CurLine++;
    if (CurLine == NB_LINES)	CurLine = 0;
    if (CurLine == FirstLine)
    {
        FirstLine++;
        if (FirstLine == NB_LINES)	FirstLine = 0;
    }
}

/*------------------------------------------------------------------*/

void
screenPrintListeCoups(
    unsigned long	nb_solution,
    unsigned long	see_solution_number,
    unsigned long	nb_position,
    const UCHAR*	moves,
    unsigned int	nb_moves)
{
    unsigned int i, j;
    char	s[ 60 ];

    if (CurPosition)	screenIncLine();

    sprintf( Lines[ CurLine ], "%s %ld (%s %ld)",
             langGetString( eStrSolution ), nb_solution ,
             langGetString( eStrPosition ), nb_position );
    for ( i = strlen( Lines[ CurLine ] ); i < LINE_SIZE - 1; i++ )
    {
        Lines[ CurLine ][ i ] = ' ';
    }

    screenIncLine();

    j = 0;
    for ( i = 0; i < nb_moves; i++ )
    {
        if ((i & 1) == 0)
        {
            sprintf( s, "%2d.", ++j );
            strcpy( &Lines[ CurLine ][ CurPosition ], s );
            CurPosition += 3;
        }

        (void)move_to_string( &moves[ i * 6 ], s );

        strcpy( &Lines[CurLine ][ CurPosition ], s );
        CurPosition += strlen( s );

        if ( CurPosition >= LINE_SIZE - 1)	screenIncLine();
    }

    if ( see_solution_number != 0 )
    {
        if (CurPosition > LINE_SIZE - 12)
        {
            strcpy( &Lines[ CurLine ][ CurPosition ], "...     ");
            screenIncLine();
            sprintf( Lines[ CurLine ], "%s %ld",
                     langGetString( eStrSeeSolution ), see_solution_number );
            CurPosition = 15;
        }
        else
        {
            sprintf( &Lines[ CurLine ][ CurPosition ], "... %s %ld",
                    langGetString( eStrSeeSolution ), see_solution_number );
        }
    }

    for ( i = strlen( Lines[ CurLine ] ); i < LINE_SIZE - 1; i++ )
    {
        Lines[ CurLine ][ i ] = ' ';
    }

    screenPrintLines();
}

void
screenEnd(const char* begin,
          const char* end,
          const char* stopped,
          const char* time)
{
    int		i;
    char*	s = "                                                                                ";
    char*	s2 = "         ";

    move(0, 9);

    attrset(ATTR_WHITE_ON_BLACK);
    for (i=0; i<16; i++)
    {
        mvaddstr(9 + i, 0, s);
    }

    mvaddstr(5, 8, s2);
    mvaddstr(6, 8, s2);

    mvaddstr(10, 0, begin);
    mvaddstr(11, 0, end);
    mvaddstr(12, 0, stopped);
    mvaddstr(13, 0, time);

    refresh();
    curs_set( 1 );		/* Reset cursor back to normal */
    endwin();
}

static void
screenShowCurrentMoves(
    const UCHAR* moves,
    unsigned int nb_moves)
{
    static char change_indicator = ' ';
    char one_move[ 10 ];
    unsigned int i;

    move(7, 17);
    for ( i = 0; i < nb_moves &&  i < 15; i++ )
    {
        attrset((i & 1) == 0 ? ATTR_WHITE_ON_BLACK : ATTR_BROWN_ON_BLACK);
        addstr(move_to_short_string( &moves[ i * 6], one_move ));
    }
    for (; i < 15; i++ )
    {
        addstr("    ");
    }

    move(8, 17);
    for (; i < nb_moves && i < 30; i++ )
    {
        attrset((i & 1) == 0 ? ATTR_WHITE_ON_BLACK : ATTR_BROWN_ON_BLACK);
        addstr(move_to_short_string( &moves[ i * 6], one_move ));
    }
    for (; i < 30; i++ )
    {
        addstr("    ");
    }

    change_indicator = ' ' + '*' - change_indicator;
    attrset(ATTR_WHITE_ON_BLACK);
    addch(change_indicator);
}

bool_t
screenKeyboard(
    bool_t		skipOK,
    const UCHAR*	moves,
    unsigned int	nb_moves,
    unsigned long	nb_positions )
{
    int			key;
    const char* const	stopTxt = langGetString(eStrStop);
    const char* const	skipTxt = langGetString(eStrSkip);
    const char*		texte;
    const char* const	empty = "         ";
    bool_t		skip;

    key = Screen_Key;
    while (getch() != ERR);		/* Flush input */

    if (key == KEY_RESIZE)
    {
        int row, col;
        getmaxyx(stdscr, row, col);
        screenSetSize(row, col);
        screenWriteImmobileAssoc();
        screenPrintLines();
        return FALSE;
    }

    key = tolower(key);

    if (key == ' ')
    {
        screenShowCurrentMoves(moves, nb_moves);
        return FALSE;
    }

    if ( skipOK && key == 's' )
    {
        texte = skipTxt;
        skip = TRUE;
    }
    else
    {
        texte = stopTxt;
        skip = FALSE;
    }

    attrset(ATTR_RED_ON_BLACK);
    mvaddstr(STOP_Y, STOP_X, texte);

    refresh();
    while ((key = getch()) == ERR);

    key = tolower( key );

    attrset(ATTR_WHITE_ON_BLACK);
    mvaddstr(STOP_Y, STOP_X, empty);

    refresh();

    if ( key == langGetChar( eCharYes ) )
    {
        if ( skip )
        {
            return TRUE ;
        }
        else
        {
            mainEnd( eStrStoppedByUser, nb_positions );
            if ( skipOK )	hashFree();
            exit( EXIT_SUCCESS );
        }
    }

    return FALSE;
}

#endif /* HAVE_LIBNCURSES */
