/*
 * test_moves_order_pawns.c --- Order the moves of the current position.
 *
 * Copyright (c) 2015, 2016 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of Natch.
 *
 * Natch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Natch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include        "common.h"

#include	"test_moves_order_pawns.h"
#include	"test_moves_order_tools.h"
#include	"pcpj.h"
#include	"distance.h"
#include	"pcpjtool.h"

/*--------------------------------------------------------------------------*/

typedef enum
{
    e_Pawn_Waits_Before_Moving,
    e_Piece_Blocks_With_First_Move,
    e_Piece_Blocks_With_Last_Move,
    e_Pawn_Last_Move_Captured,
    e_Pawn_Captured
} pawn_constraint_type_t ;

/*--------------------------------------------------------------------------*/

bool_t
isSameDirection(square_t from1, square_t to1, square_t from2, square_t to2)
{
    int d1, d2;
    int dx1 = column(to1) - column(from1);

    if (dx1 == 0)
    {
        d1 = row(to1) - row(from1);
        d2 = row(to2) - row(from2);
    }
    else
    {
        d1 = dx1;
        d2 = column(to2) - column(from2);
    }
    return (d1 > 0 && d2 > 0) || (d1 < 0 && d2 < 0);
}

/*--------------------------------------------------------------------------*/

static
bool_t
add_pawn_constraint(
    struct moves_order_t* moves_order,
    const exploration_t* exploration,
    twin_squares_t* twin_squares,
    pawn_constraint_type_t type,
    const piece_t* pawn)
{
    distance_t distance_table[6 * SIZE_PIECE];
    unsigned int i;

    distance_create_table(distance_table, twin_squares);
    distance_push(distance_table);

    for (i = INDEX_ROI_BLANC ; i <= INDEX_PION_NOIR_H ; i++)
    {
        const piece_t* piece = &exploration->pieces[i];
        unsigned int move_index = 0;
        piece_type_t piece_type = piece->typePiece;

        if (piece->destinations[0].distance == 0
            || piece == pawn
            || piece_type == CAVALIER
            || (piece_type == PION
                && (!has_promoted(piece) || piece->typePromotion == CAVALIER)))
            continue;

        if (type == e_Piece_Blocks_With_Last_Move
            && pawn->destinations[pawn->nbDestination - 2].capturedPiece == i)
        {
            /* The captured piece is not blocked by the pawn. */
            continue;
        }

        if (piece_type == PION)
        {
            bool_t ok = TRUE;
            while (piece->destinations[move_index].destination
                   != piece->casePromotion)
            {
                if (type == e_Pawn_Captured
                    && (piece->destinations[move_index].capturedPiece
                        == pawn->index))
                {
                    /* The piece that captures the pawn is no not blocked after
                     * the capture. */
                    ok = FALSE;
                    break;
                }
                move_index++;
            }
            if (!ok) continue;
            if (piece->typePromotion != PION) {
                piece_type = piece->typePromotion;
                move_index++;
            }
        }
        else if (piece_type == ROI && has_castled(piece))
        {
            /* We skip the castling move. */
            move_index = 1;
        }

        while (move_index < piece->nbDestination)
        {
            square_t current_square = move_index == 0
                ? piece->caseInitiale
                : piece->destinations[move_index - 1].destination;

            if (type == e_Pawn_Captured
                && (piece->destinations[move_index].destination
                    == pawn->caseArrivee)
                && (has_captured(piece, move_index)))
            {
                /* The pieces that make captures on the square where the pawn is
                 * captured, are no more concerned after the capture. */
                break;
            }

            if (type == e_Pawn_Waits_Before_Moving
                && piece->destinations[move_index].distance > 1)
            {
                /* if e_Pawn_Waits_Before_Moving is tested, we eliminate the
                 * case Ra1xa7(2) and a2-a4 with white having a spare move
                 * allowing to play a2-a3. */
                move_index++;
                continue;
            }

            if (current_square != CASE_ARRIVEE_QUELCONQUE
                && (piece->destinations[move_index].destination
                    != CASE_ARRIVEE_QUELCONQUE))
            {
                distance_t delta = distance_delta(
                    piece_type,
                    piece->camp,
                    current_square,
                    piece->destinations[move_index].destination);
                if (delta > 0)
                {
                    if ((piece->camp == BLANC
                         && delta > NbCoupsBlancsRestants)
                        || (piece->camp == NOIR
                            && delta > NbCoupsNoirsRestants))
                    {
                        bool_t ok;
                        switch(type)
                        {
                        case e_Pawn_Waits_Before_Moving:
                        {
                            piece_index_t before_piece_index = i;
                            ok = TRUE;
                            if (!isSameDirection(
                                    pawn->caseInitiale,
                                    pawn->destinations[0].destination,
                                    current_square,
                                    piece->destinations[move_index].destination))
                            {
                                if (move_index + 1 < piece->nbDestination)
                                {
                                    move_index++;
                                }
                                else if (is_captured(piece))
                                {
                                    find_move_after_capture(
                                        i,
                                        exploration,
                                        &before_piece_index,
                                        &move_index);
                                }
                                else
                                {
                                    ok = FALSE;
                                    if (shouldPrintEliminated())
                                    {
                                        print_eliminated_for_move_order(
                                            moves_order,
                                            before_piece_index,
                                            move_index,
                                            e_First_Move,
                                            pawn->index,
                                            0,
                                            e_First_Move,
                                            exploration,
                                            "pawn waits before moving");
                                    }
                                }
                            }

                            if (ok)
                            {
                                ok = add_one_order(
                                    moves_order,
                                    before_piece_index,
                                    move_index,
                                    e_First_Move,
                                    pawn->index,
                                    0,
                                    e_First_Move,
                                    exploration,
                                    "pawn waits before moving");
                            }
                        }
                        break;
                        case e_Piece_Blocks_With_First_Move:
                        {
                            ok = add_one_order(
                                moves_order,
                                pawn->index,
                                1,
                                e_First_Move,
                                i,
                                move_index,
                                e_Last_Move,
                                exploration,
                                "piece must evacuate line first");
                        }
                        break;
                        case e_Piece_Blocks_With_Last_Move:
                        {
                            piece_index_t after_piece_index;
                            unsigned int after_move_index;
                            if (pawn->typePiece == PION)
                            {
                                piece_index_t captured =
                                    pawn->destinations[
                                        pawn->nbDestination - 2].capturedPiece;
                                after_piece_index = captured;
                                after_move_index =
                                    exploration->pieces[captured].nbDestination
                                    - 1;
                            }
                            else
                            {
                                after_piece_index = pawn->index;
                                after_move_index = pawn->nbDestination - 2;
                            }
                            ok = add_one_order(
                                moves_order,
                                i,
                                move_index,
                                e_Last_Move,
                                after_piece_index,
                                after_move_index,
                                e_Last_Move,
                                exploration,
                                "piece waits before blocking line on last move");
                        }
                        break;
                        case e_Pawn_Last_Move_Captured:
                        {
                            ok = add_one_order(
                                moves_order,
                                i,
                                move_index,
                                e_First_Move,
                                pawn->index,
                                pawn->nbDestination - 2,
                                e_Last_Move,
                                exploration,
                                "pawn waits before last capture");
                        }
                        break;
                        case e_Pawn_Captured:
                        {
                            piece_index_t before_piece_index;
                            unsigned int before_move_index;
                            find_move_after_capture(
                                pawn->index,
                                exploration,
                                &before_piece_index,
                                &before_move_index);

                            /* It may happen that the final capturing piece is
                             * the tested piece itself. */
                            if (before_piece_index == i
                                && before_move_index == move_index)
                            {
                                ok = TRUE;
                            }
                            else
                            {
                                ok = add_one_order(
                                    moves_order,
                                    before_piece_index,
                                    before_move_index,
                                    e_First_Move,
                                    i,
                                    move_index,
                                    e_Last_Move,
                                    exploration,
                                    "pawn must be captured first");
                            }
                        }
                        break;
                        }

                        if (!ok)
                        {
                            distance_pop();
                            return FALSE;
                        }
                        break;
                    }
                }
            }

            move_index++;
        }
    }

    distance_pop();
    return TRUE;
}

bool_t
add_pawn_line_constraints(
    struct moves_order_t*	moves_order,
    const exploration_t*	exploration,
    twin_squares_t*             twin_squares)
{
    piece_index_t i;

    for (i = INDEX_ROI_BLANC; i <= INDEX_PION_NOIR_H; i++)
    {
        const piece_t* p = &exploration->pieces[i];
        int deltaY = row(p->caseArrivee) - row(p->caseInitiale);
        unsigned int nbDest = p->nbDestination;
        bool_t ok = TRUE;
        int nb_spare_moves =
            p->camp == BLANC ? NbCoupsBlancsRestants : NbCoupsNoirsRestants;

        if (is_knight(p)) continue;

        if (!is_pawn(p))
        {
            if (nbDest > 1
                && p->destinations[nbDest - 1].distance == 1
                && !is_captured(p)
                // If nb_spare_moves > 0, the piece can open the line, even with
                // only one spare move is left.  Example: Ra1xPa7-a8, the rook
                // can return to a1 to open the line a8-a2 to the bRa8.
                && nb_spare_moves == 0
                && (p->destinations[nbDest - 2].destination
                    != CASE_ARRIVEE_QUELCONQUE))
            {
                square_t current_square= p->destinations[nbDest - 2].destination;

                /* The piece type is a pawn, so all trajectories are checked,
                 * even if, for example, for rooks if the twin square is also
                 * for a rook.
                 */
                add_twin_square(
                    twin_squares,
                    current_square,
                    p->caseArrivee,
                    twin_direction(current_square, p->caseArrivee),
                    PION,
                    p->camp,
                    p->caseInitiale);

                ok = add_pawn_constraint(
                    moves_order,
                    exploration,
                    twin_squares,
                    e_Piece_Blocks_With_Last_Move,
                    p);

                remove_last_twin_square(twin_squares);

                if (!ok) return FALSE;
            }

            if (nbDest > 1
                && p->destinations[0].distance == 1
                && !(is_rook(p) && has_castled(p))
                && (nb_spare_moves == 0
                    || (nb_spare_moves == 1 && (is_rook(p) || is_bishop(p))))
                && (p->destinations[0].destination != CASE_ARRIVEE_QUELCONQUE))
            {
                square_t current_square = p->caseInitiale;
                square_t dest = p->destinations[0].destination;

                /* The piece type is a pawn, so all trajectories are checked,
                 * even if, for example, for rooks ef the twin square is also
                 * for a rook.
                 */
                add_twin_square(
                    twin_squares,
                    current_square,
                    dest,
                    twin_direction(current_square, dest),
                    PION,
                    p->camp,
                    p->caseInitiale);

                ok = add_pawn_constraint(
                    moves_order,
                    exploration,
                    twin_squares,
                    e_Piece_Blocks_With_First_Move,
                    p);

                remove_last_twin_square(twin_squares);

                if (!ok) return FALSE;
            }

            continue;
        }

        /* Only pawns after this line. */

        if (nbDest > 1 && !has_captured(p, 0))
        {
            add_twin_square(
                twin_squares,
                p->caseInitiale,
                p->destinations[0].destination,
                e_vertical,
                PION,
                p->camp,
                p->caseInitiale);

            ok = add_pawn_constraint(
                moves_order,
                exploration,
                twin_squares,
                e_Piece_Blocks_With_First_Move,
                p);

            remove_last_twin_square(twin_squares);

            if (!ok) return FALSE;
        }

        if (nbDest > 1 && !has_promoted(p) && !is_captured(p))
        {
            int depX = column(p->caseArrivee);
            int depY = row(p->caseArrivee) >> 4;
            int arrX = column(p->destinations[nbDest - 2].destination);
            int arrY = row(p->destinations[nbDest - 2].destination) >> 4;

            twin_direction_t direction =
                !has_captured(p, nbDest - 1)
                ? e_vertical
                : ((depX - arrX) == (depY - arrY) ? e_up_right : e_up_left);

            add_twin_square(
                twin_squares,
                p->caseArrivee,
                p->destinations[nbDest - 2].destination,
                direction,
                PION,
                p->camp,
                p->caseInitiale);

            ok = add_pawn_constraint(
                moves_order,
                exploration,
                twin_squares,
                (!has_captured(p, nbDest - 1)
                 ? e_Piece_Blocks_With_Last_Move
                 : e_Pawn_Last_Move_Captured),
                p);

            remove_last_twin_square(twin_squares);

            if (!ok) return FALSE;
        }

        if (nbDest == 1 && is_captured(p) && !p->has_never_moved)
        {
            add_twin_square(
                twin_squares,
                p->caseInitiale,
                p->destinations[0].destination,
                e_vertical,
                PION,
                p->camp,
                p->caseInitiale);

            ok = add_pawn_constraint(
                moves_order,
                exploration,
                twin_squares,
                e_Pawn_Captured,
                p);

            remove_last_twin_square(twin_squares);

            if (!ok) return FALSE;
        }

        if (nbDest == 1
            && !is_captured(p)
            && deltaY != 0
            && deltaY != 16
            && deltaY != -16
            && (deltaY != 32 || NbCoupsBlancsRestants > 0)
            && (deltaY != -32 || NbCoupsNoirsRestants > 0))
        {
            int increment = p->camp == BLANC
                ? (NbCoupsBlancsRestants > 0 ? 16 : 32)
                : (NbCoupsNoirsRestants > 0 ? -16 : -32);

            replace_twin_square(
                twin_squares,
                p->caseInitiale,
                p->caseInitiale + increment,
                p->caseArrivee);

            ok = add_pawn_constraint(
                moves_order,
                exploration,
                twin_squares,
                e_Pawn_Waits_Before_Moving,
                p);

            replace_twin_square(
                twin_squares,
                p->caseInitiale,
                p->caseInitiale,
                p->caseArrivee);

            if (!ok) return FALSE;
        }
    }

    return TRUE;
}
